require 'uu/os/cmd/command_client'
require 'uu/os/security/session'

module UU
  module C3

    # Client of service providing operations for self-service deployment of uuApps.
    class AppDeployment

      # Remote service path.
      # @private
      PATH = 'uu-c3/AppDeployment'

      # Default URLs where common commands are running.
      # @private
      DEFAULT_COMMON_BASE_URL = /https:\/\/(api|cmd)\.plus4u\.net/

      # Default URL for C3 (and other syscalls).
      # @private
      DEFAULT_C3_BASE_URL = 'https://sys.plus4u.net'

      # Try to auto-configure client base URL.
      ext_config = UU::OS::CMD::CommandClient.const_get(:EXTERNAL_CONFIGURATION)
      base_url = ext_config.get(:base_url, PATH)
      if base_url.nil? || base_url =~ DEFAULT_COMMON_BASE_URL
        UU::OS::CMD::CommandClient.load_configuration({'uu.c3.base_url' => DEFAULT_C3_BASE_URL})
      end
      # Command is not yet invoked through uuGateway, therefore do not append routing data.
      UU::OS::CMD::CommandClient.load_configuration({'uu.c3.append_routing_data' => 'false'})

      # Creates a new instance of +AppDeployment+.
      #
      # @param session [UU::OS::Security::Session] Instance of session representing authenticated user
      #   (can be obtained with {UU::OS::Security::Session#current_session} method).
      def initialize(session = nil)
        @client = UU::OS::CMD::CommandClient.new(PATH, session)
      end

      # Deploys the specified appbox to the given pool.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   future = app_deployment.deploy("ues:DEV1234-BT:DEV",
      #     appBoxUri: "ues:DEV1234-BT:CDS.GB_1.0.6",
      #     msgBusMapping: {"MESSAGEBUS" => "ues:OT-BT:MSGBUS_042"},
      #     uuEEs: ["1-2345-6"])
      #   future.get # blocking operation
      #
      # @example Configuring uuApp at deployment time
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   future = app_deployment.deploy("ues:DEV1234-BT:DEV",
      #     appBoxUri: "ues:DEV1234-BT:CDS.GB_1.0.6",
      #     config: {
      #         param1: "value1",
      #         param2: "osid:value2", # uuAppObjectStore ID (hexa32Code; needs to be whitelisted for the uuSubApp)
      #         param3: "bsid:value3", # uuAppBinaryStore ID (hexa32Code; needs to be whitelisted for the uuSubApp)
      #         redis_uri: "redis://:secrets@example.com:1234/9"
      #     }
      #   )
      #   app_deployment_uri = future.get # blocking operation
      #
      # @example Reading deployment-time configuration (uuApps 2.1)
      #   # read config everywhere inside application
      #   redis_uri = UU::OS::Application::Config[:redis_uri] # String "redis://:secrets@example.com:1234/9"
      #
      #   # You can set default values inside application in file project_root/config/app.rb, example:
      #
      #   # cds-gb/config/app.rb:
      #
      #   # default Redis uri used if no config parameter is set through +deploy+ method. E.g. for local development.
      #   UU::OS::Application::Config[:redis_uri] = "redis://:secrets@localhost:1234/9"
      #
      #
      # @param resource_pool_uri [UU::OS::UESURI, String] Target pool to deploy appbox to.
      #   This parameter is required.
      # @param parameters [Hash] AppDeployment parameters.
      # @option parameters [UU::OS::UESURI, String] appBoxUri The appbox to deploy.
      #   This parameter is required.
      # @option parameters [Array<String>] uuEEs Optional list of uuEEs
      #   (identified by +4U ID) to use by the deployed application in the
      #   territory specified by +resource_pool_uri+. May be +nil+.
      #   If this parameter is specified, the list is validated against
      #   whitelist configured in UU Cloud metadata (by UU Cloud provider) -
      #   only uuEEs specified in the whitelist may be used by the app in the
      #   target territory (identified by +resource_pool_uri+); if this
      #   parameter contains uuEEs not specified in the whitelist, the command
      #   fails with an exception (since uuC3 1.2.4).
      #   If this parameter is not specified, the list of uuEEs is obtained
      #   from the whitelist stored in UU Cloud metadata, if specified for the
      #   given app and territory (since uuC3 1.2.4).
      # @option parameters [Hash<String, String>] msgBusMapping In case message
      #   bus usage is declared in application deployment descriptors (which is
      #   necessary for asynchronous commands to work), mapping from alias
      #   message bus code to actual message bus to use must be provided. It is
      #   done via simple key-value hash, where key is message bus code used in
      #   deployment descriptors and value is UESURI of an existing message bus.
      #   May be +nil+.
      # @option parameters [Hash<String, Object>] config Optional application configuration.
      #   Configuration is readable everywhere across the deployed application through
      #   {UU::OS::Application::Config} class. Typically is used for production runtime configuration like
      #   resources URI (URI to uuAppObjectStore, uuAppBinaryStore, Redis etc.) Parameter may contain only uuJson serializable data
      #   (as described in UNI-BT:44191583772172834). It means that only basic values are supported
      #   (String, Fixnum, Boolean, Hash, Array), others object are serialized to String. For example if you use ruby Time
      #   class as parameter, so String "2016-02-15 13:54:33 +0100" (serialized Time) will be available in application.
      # @option parameters [TrueClass, FalseClass] forceBuild Flag determining whether
      #   to force building new uuNode images (regardless whether they exist for the
      #   given uuApp Box version or not) or to reuse existing uuNode images (if they
      #   exist). Default behaviour is to reuse existing uuNode images (i.e. +false+
      #   value).
      #
      #   * +true+ means that any existing uuNode images for this uuApp Box version
      #     are ignored and new ones get built anew.
      #     * Positive effect: uuNodes get deployed with images built from uuApp Packs
      #       attached to the uuApp Box being deployed.
      #     * Negative effect: The deployment process takes more time because uuNode
      #       images need to be built afresh (even if they already exist).
      #   * +false+ (default value) means that uuNode images get built only if they
      #     don't exist for the given uuApp Box version).
      #     * Positive effect: The deployment process may take less time if the uuNode
      #       images already exist.
      #
      #   It is recommended to not set this flag if you are not sure. May be +nil+.
      #
      #   Note that you cannot set +forceBuild+ for currently deployed applications.
      #   If you need to rebuild uuNode images for deployed application you have
      #   two options:
      #   * undeploy the application first and then deploy with +forceBuild+
      #   * change application version
      #   The reason is that there cannot exist two uuNode images for the same version
      #   of the same application.
      #
      #
      # @option parameters [String] asid AppServer identifier. May be +nil+ but an
      #   AppServer with the specified ASID must exist if specified. Only
      #   applicable to uuAppsg01 (UAF uuApps).
      # @option parameters [TrueClass, FalseClass] returnFuture Flag
      #   determining whether to return +future+ (+true+) that is used to wait
      #   for the end of the asynchronous part of the deploy or to return
      #   +appDeploymentUri+ instead (+false+). May be +nil+.
      #   Default value is +true+ to preserve backward compatibility.
      # @option parameters [UU::C3::AppDeploymentSlot] targetSlot Slot used for AppDeployment.
      #   Slot defines how will be application accessible.
      #   Point of slots is enable testing on production environment and following swap on the production gateway.(Principe of Blue/Green Deployment)
      #   For classic deploy and access application over standard gateway is defined slot 'PRODUCTION'(default value).
      #   For testing in production environment is defined slot 'BETA'.
      #
      #   Note: Configuration used in 'BETA' slot will be after swap used in 'PRODUCTION' slot.
      #
      # @return [UU::OS::Lang::Future, UU::OS::UESURI] Either instance of future to use to wait for
      #   the deployment operation to finish (Note that maximum timeout of future
      #   is set to 30 seconds, which is unfortunately insufficient time to finish
      #   deployment. Therefore timeout error might be expected when +Future.get+
      #   is invoked immediately after the command invocation. It is however
      #   possible to wait repeatedly using the same instance of the returned
      #   +Future+ object. +Future.get+ returns +appDeploymentUri+ usable e.g. for
      #   the {#undeploy} method.) or +appDeploymentUri+. See parameter returnFuture.
      # @see .undeploy
      def deploy(resource_pool_uri, parameters = {})
        result = @client.invoke(:deploy, resource_pool_uri, parameters: parameters)
        return UU::OS::UESURI.new(result) if result.is_a?(String)
        result
      end

      # Deploys the specified appbox to the given pool. This method is a convenience
      # shortcut for the {#deploy} instance method.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   future = UU::C3::AppDeployment.deploy("ues:DEV1234-BT:DEV",
      #     appBoxUri: "ues:DEV1234-BT:CDS.GB_1.0.6",
      #     msgBusMapping: {"MESSAGEBUS" => "ues:OT-BT:MSGBUS_042"},
      #     uuEEs: ["1-2345-6"])
      #   future.get # blocking operation
      #
      # @example Configuring uuApp at deployment time
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   future = UU::C3::AppDeployment.deploy("ues:DEV1234-BT:DEV",
      #     config: {
      #         param1: "value1",
      #         param2: "osid:value2", # uuAppObjectStore ID (hexa32Code; needs to be whitelisted for the uuSubApp)
      #         param3: "bsid:value3", # uuAppBinaryStore ID (hexa32Code; needs to be whitelisted for the uuSubApp)
      #         redis_uri: "redis://:secrets@example.com:1234/9"
      #     }
      #   )
      #   app_deployment_uri = future.get # blocking operation
      #
      # @example Reading deployment-time configuration (uuApps 2.1)
      #   # read config everywhere inside application
      #   redis_uri = UU::OS::Application::Config[:redis_uri] # String "redis://:secrets@example.com:1234/9"
      #
      #   # You can set default values inside application in file project_root/config/app.rb, example:
      #
      #   # cds-gb/config/app.rb:
      #
      #   # default Redis uri used if no config parameter is set through +deploy+ method. E.g. for local development.
      #   UU::OS::Application::Config[:redis_uri] = "redis://:secrets@localhost:1234/9"
      #
      # @return [UU::OS::Lang::Future] Instance of future to use to wait for
      #   the deployment operation to finish.
      def self.deploy(resource_pool_uri, parameters = {})
        self.new(UU::OS::Security::Session.current_session).deploy(resource_pool_uri, parameters)
      end

      # Undeploys the specified uuApp from the pool and releases all related
      # allocated resources. This command can be invoked repeatedly with the
      # same parameters, to ensure full cleanup even after unexpected failures from
      # previous invocations.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   future = app_deployment.undeploy("ues:[123]:[56a3b8ec69a1b474a0000086]:[56a3b8ec69a1b474a0000089]")
      #   future.get # blocking operation
      #
      # @param app_deployment_uri [UU::OS::UESURI, String] URI of deployment which will by undeployed.
      # @param parameters [Hash] Undeploy parameters.
      # @option parameters [UU::C3::AppDeploymentSlot] targetSlot Slot to undeploy. Default value is 'PRODUCTION'.
      # @return [UU::OS::Lang::Future, nil] Instance of future to use to wait for
      #   the undeployment operation to finish. Note that maximum timeout of future
      #   is set to 30 seconds, which is unfortunately insufficient time to finish
      #   undeployment. Therefore timeout error might be expected when +Future.get+
      #   is invoked immediately after the command invocation. It is however
      #   possible to wait repeatedly using the same instance of the returned
      #   +Future+ object. Returns nil if no deployed resources were found.
      # @see .undeploy
      def undeploy(app_deployment_uri, parameters = {})
        @client.invoke(:undeploy, app_deployment_uri, parameters: parameters)
      end

      # Undeploys the specified uuApp from the pool and releases all related
      # allocated resources. This method is a convenience shortcut for the
      # {#undeploy} instance method.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   future = app_deployment.undeploy("ues:[123]:[56a3b8ec69a1b474a0000086]:[56a3b8ec69a1b474a0000089]")
      #   future.get # blocking operation
      #
      # @return [UU::OS::Lang::Future, nil] Instance of future to use to wait for
      #   the undeployment operation to finish. Returns nil if no deployed resources were found.
      def self.undeploy(app_deployment_uri, parameters = {})
        self.new(UU::OS::Security::Session.current_session).undeploy(app_deployment_uri, parameters)
      end


      def redeploy(app_deployment_uri, parameters = {})
        @client.invoke(:redeploy, app_deployment_uri, parameters: parameters)
      end

      # @!method redeploy(app_deployment_uri)
      #   Redeploys the specified app deployment.
      #   @example
      #     require "uu_c3"
      #
      #     UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #     app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #     future = app_deployment.redeploy("ues:[123]:[56a3b8ec69a1b474a0000086]:[56a3b8ec69a1b474a0000089]")
      #     app_deplyment_uri = future.get # blocking operation
      #   @param app_deployment_uri [UU::OS::UESURI, String] URI of deployment which will by redeployed.
      #   @return [UU::OS::Lang::Future] Instance of future to use to wait for
      #     the redeployment operation to finish. Note that maximum timeout of future
      #     is set to 30 seconds, which is unfortunately insufficient time to finish
      #     redeployment. Therefore timeout error might be expected when +Future.get+
      #     is invoked immediately after the command invocation. It is however
      #     possible to wait repeatedly using the same instance of the returned
      #     +Future+ object. +Future.get+ returns +appDeploymentUri+.

      def self.redeploy(app_deployment_uri, parameters = {})
        self.new(UU::OS::Security::Session.current_session).redeploy(app_deployment_uri, parameters)
      end

      # @!method self.redeploy(app_deployment_uri)
      #   Redeploys the specified app deployment. This method is a convenience shortcut for the
      #   {#redeploy} instance method.
      #   @example
      #     require "uu_c3"
      #
      #     UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #     future = UU::C3::AppDeployment.redeploy("ues:[123]:[56a3b8ec69a1b474a0000086]:[56a3b8ec69a1b474a0000089]")
      #     app_deplyment_uri = future.get # blocking operation
      #   @return [UU::OS::Lang::Future] Instance of future to use to wait for
      #     the redeployment operation to finish.


      # Returns capacity overview for the given pool.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   capacity = app_deployment.get_resource_pool_capacity("ues:DEV1234-BT:DEV")
      #
      # @example Response
      #   {
      #     :uri => "ues:DEV0116-BT[84723967990075193]:DEV[58bd2f89ed26b7cf63b77088]",
      #     :name => "DEV0116-BT Development Pool",
      #     :accessGroupUri => "ues:DEV0116-BT[84723967990075193]:DEV~DEPLOYERS[93947438642013117]",
      #     :auditorsUri => "ues:DEV0116-BT[84723967990075193]:DEV~LOGREADERS[64294013174317893]",
      #     :monitoringBaseUrl => "https://telemetry.plus4u.net/uu-cloudg01-monitoring/78462435-4036e1c5896443619087eb5bde9d2b70/",
      #     :contractedCapacity => {
      #       :cpu => 5,
      #       :mem => 5000,
      #       :storage => 5000
      #     },
      #     :allocatedCapacity => {
      #       :cpu => 0,
      #       :mem => 0,
      #       :storage => 0
      #     }
      #   }
      #
      # @param resource_pool_uri [UU::OS::UESURI, String] URI of the pool for which to
      #   get capacity.
      #
      # @return [Hash<Symbol, Object>] Capacity overview. Total contracted capacity
      #   (under +:contractedCapacity+ key) and currently allocated capacity
      #   (under +:allocatedCapacity+ key) are returned. Also returns
      #   ResourcePool name (under +:name+ key), AccessGroup URI
      #   (under +:accessGroupUri+ key), Auditors URI if set (under
      #   +:auditorsUri+ key) and URL of associated
      #   monitoring application if deployed (under +:monitoringBaseUrl+ key).
      # @see .get_resource_pool_capacity
      def get_resource_pool_capacity(resource_pool_uri)
        @client.invoke(:getResourcePoolCapacity, resource_pool_uri)
      end

      # Returns capacity overview for the given pool. This method is a
      # convenience shortcut for the {#get_resource_pool_capacity} instance method.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   capacity = UU::C3::AppDeployment.get_resource_pool_capacity("ues:DEV1234-BT:DEV")
      #
      # @example Response
      #   {
      #     :uri => "ues:DEV0116-BT[84723967990075193]:DEV[58bd2f89ed26b7cf63b77088]",
      #     :name => "DEV0116-BT Development Pool",
      #     :accessGroupUri => "ues:DEV0116-BT[84723967990075193]:DEV~DEPLOYERS[93947438642013117]",
      #     :auditorsUri => "ues:DEV0116-BT[84723967990075193]:DEV~LOGREADERS[64294013174317893]",
      #     :monitoringBaseUrl => "https://telemetry.plus4u.net/uu-cloudg01-monitoring/78462435-4036e1c5896443619087eb5bde9d2b70/",
      #     :contractedCapacity => {
      #       :cpu => 5,
      #       :mem => 5000,
      #       :storage => 5000
      #     },
      #     :allocatedCapacity => {
      #       :cpu => 0,
      #       :mem => 0,
      #       :storage => 0
      #     }
      #   }
      #
      # @param resource_pool_uri [UU::OS::UESURI, String] URI of the pool for which to
      #   get capacity.
      #
      # @return [Hash<Symbol, Object>] Capacity overview. Total contracted capacity
      #   (under +:contractedCapacity+ key) and currently allocated capacity
      #   (under +:allocatedCapacity+ key) are returned. Also returns
      #   ResourcePool name (under +:name+ key), AccessGroup URI
      #   (under +:accessGroupUri+ key), Auditors URI if set (under
      #   +:auditorsUri+ key) and URL of associated
      #   monitoring application if deployed (under +:monitoringBaseUrl+ key).
      def self.get_resource_pool_capacity(resource_pool_uri)
        self.new(UU::OS::Security::Session.current_session).get_resource_pool_capacity(resource_pool_uri)
      end

      # Returns list of applications deployed in the given pool.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   apps = app_deployment.get_app_deployment_list("ues:DEV1234-BT:DEV", :state => UU::C3::AppDeploymentState::DEPLOYED)
      #
      # @example Response
      #  {
      #      :totalSize => 1300,
      #      :criteria => {
      #          :pageIndex => 2,
      #          :pageSize => 30
      #      },
      #      :pageEntries => {
      #          :appBoxUri => "ues:[123]:[456]",
      #          :asid => "df590bef9e834400bc379185d2a43116",
      #          :appBoxCode => "FOO_ART",
      #          :code => "CDS.GB",
      #          :state => "DEPLOYED",
      #          :version => "1.0.84",
      #          :urlPath => "/cds-gb",
      #          :multitenant => true,
      #          :config => {
      #              code => "CDS.GB",
      #              :version => "1.0.84",
      #              :urlPath => "/cds-gb"
      #          },
      #          :resourcePoolUri => "ues:TERRITORY[123]:POOL[59955e0d123ddbdfe046c6f8]",
      #          :tenantUri => "ues:OT-BT[0]:TERRITORY[59955e0d123ddb574146c6f6]",
      #          :lastStateChange => "2017-08-17T09:12:54.052Z",
      #          :uri => "ues:TERRITORY[123]:POOL[59955e0d123ddbdfe046c6f8]:CDS.GB[59955e0e123ddb387646c6fc]",
      #          :nodeSets => [
      #              {
      #                  :uri => "ues:TERRITORY[99923616732452117]:CDS_GB_VUC_PACK/CDS_GB_VUC_CONTAINER/CDS_BT[5739c55927928929b914820f]",
      #                  :code => "CDS_GB_VUC_PACK/CDS_GB_VUC_CONTAINER/CDS_BT",
      #                  :nodeCount => 1
      #              },
      #              {
      #                  :uri => "ues:TERRITORY[99923616732452117]:CDS_GB_CMD_PACK/CDS_GB_CMD_CONTAINER/CDS_BT[5739c55927928929b914820f]",
      #                  :code => "CDS_GB_CMD_PACK/CDS_GB_CMD_CONTAINER/CDS_BT",
      #                  :nodeCount => 1
      #              }
      #          ],
      #          :allocatedCapacity => {
      #              :cpu => 2100,
      #              :mem => 2100,
      #              :storage => 2100
      #          },
      #          :msgBusses => [
      #              {
      #                  :msgBusUri => "ues:SYSTEM:MSGBUS01",
      #                  :queuePairsCount => 2
      #              },
      #              {
      #                  :msgBusUri => "ues:SYSTEM:MSGBUS02",
      #                  :queuePairsCount => 1
      #              }
      #          ]
      #      }
      #  }
      #
      # @param resource_pool_uri [UU::OS::UESURI, String] URI of the pool for which to
      #   get deployed applications list.
      # @param criteria [Hash] Optional criteria for listing deployments.
      # @option criteria [UU::C3::AppDeploymentState, String] state State of AppDeployment.
      # @option criteria [Fixnum] pageSize Size of one page of records. Default is 1000.
      # @option criteria [Fixnum] pageIndex Page index where 0 means first page, 1 second and so on. Default is 0.
      # @return [Hash<Symbol, Object>] Paged result set.
      # @see .get_app_deployment_list
      def get_app_deployment_list(resource_pool_uri, criteria = {})
        @client.invoke(:getAppDeploymentList, resource_pool_uri, :parameters => criteria)
      end

      # Returns list of applications deployed in the given pool. This method is
      # a convenience shortcut for the {#get_app_deployment_list} instance method.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   apps = UU::C3::AppDeployment.get_app_deployment_list("ues:DEV1234-BT:DEV", :state => UU::C3::AppDeploymentState::DEPLOYED)
      #
      # @return [Hash<Symbol, Object>] Paged result set.
      def self.get_app_deployment_list(resource_pool_uri, criteria = {})
        self.new(UU::OS::Security::Session.current_session).get_app_deployment_list(resource_pool_uri, criteria)
      end

      # Shares the specified app deployment to the given territories.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   future = UU::C3::AppDeployment.share("ues:DEV1234-BT:PROD:CDS.GB",
      #     territories: ["ues:MT.123:MT.123", "ues:MT.456:MT.456"]
      #   )
      #   future.get # blocking operation
      #
      # @param app_deployment_uri [UU::OS::UESURI, String] URI of app deployment which will be shared.
      # @param parameters [Hash] AppDeployment share parameters.
      # @option parameters [Array] territories List of territory URIs where application will be shared. Maximum is 1000 territories per command.
      # @option parameters [TrueClass, FalseClass] reportProgress Disable (false) or enable (true) progress reporting
      # (i.e. creation of the appbox activity). By default, reporting is enabled.
      # @return [UU::OS::Lang::Future] Instance of future to use to wait for
      #   the share operation to finish. Note that maximum timeout of future
      #   is set to 30 seconds, which is unfortunately insufficient time to finish
      #   sharing. Therefore timeout error might be expected when +Future.get+
      #   is invoked immediately after the command invocation. It is however
      #   possible to wait repeatedly using the same instance of the returned
      #   +Future+ object.
      # @see .share
      def share(app_deployment_uri, parameters = {})
        @client.invoke(:share, app_deployment_uri, parameters: parameters)
      end

      # Shares the specified app deployment to the given territories. This method is a convenience shortcut for the
      # {#share} instance method.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   UU::C3::AppDeployment.share("ues:DEV1234-BT:DEV:CDS.GB",
      #     territories: ["ues:MT.123:MT.123", "ues:MT.456:MT.456"]
      #   )
      #   future.get # blocking operation
      #
      # @param app_deployment_uri [UU::OS::UESURI, String] URI of app deployment which will be shared.
      # @param parameters [Hash] AppDeployment share parameters.
      # @option parameters [Array] territories List of territory URIs where application will be shared. Maximum is 1000 territories per command.
      # @option parameters [TrueClass, FalseClass] reportProgress Disable (false) or enable (true) progress reporting
      # (i.e. creation of the appbox activity). By default, reporting is enabled.
      # @return [UU::OS::Lang::Future] Instance of future to use to wait for
      #   the share operation to finish. Note that maximum timeout of future
      #   is set to 30 seconds, which is unfortunately insufficient time to finish
      #   sharing. Therefore timeout error might be expected when +Future.get+
      #   is invoked immediately after the command invocation. It is however
      #   possible to wait repeatedly using the same instance of the returned
      #   +Future+ object.
      def self.share(app_deployment_uri, parameters = {})
        self.new(UU::OS::Security::Session.current_session).share(app_deployment_uri, parameters)
      end

      # Terminates sharing of the specified app deployment for the given territories. Routing data for the particular
      # app deployment and territories are removed.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   future = app_deployment.unshare("ues:DEV1234-BT:DEV:CDS.GB",
      #     territories: ["ues:MT.123:MT.123", "ues:MT.456:MT.456"]
      #   )
      #   future.get # blocking operation
      #
      # @param app_deployment_uri [UU::OS::UESURI, String] URI of app deployment which will be unshared.
      # @param parameters [Hash] AppDeployment unshare parameters.
      # @option parameters [Array] territories List of territory URIs from where the app deployment will be unshared. Maximum is 1000 territories per command.
      # @option parameters [TrueClass, FalseClass] reportProgress Disable (false) or enable (true) progress reporting
      # (i.e. creation of the appbox activity). By default, reporting is enabled.
      # @return [UU::OS::Lang::Future] Instance of future to use to wait for
      #   the unshare operation to finish. Note that maximum timeout of future
      #   is set to 30 seconds, which is unfortunately insufficient time to finish
      #   unsharing. Therefore timeout error might be expected when +Future.get+
      #   is invoked immediately after the command invocation. It is however
      #   possible to wait repeatedly using the same instance of the returned
      #   +Future+ object.
      # @see .unshare
      def unshare(app_deployment_uri, parameters = {})
        @client.invoke(:unshare, app_deployment_uri, parameters: parameters)
      end

      # Terminates sharing of the specified app deployment for the given territories. This method is a convenience
      # shortcut for the {#unshare} instance method.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   future = UU::C3::AppDeployment.unshare("ues:DEV1234-BT:DEV:CDS.GB",
      #     territories: ["ues:MT.123:MT.123", "ues:MT.456:MT.456"]
      #   )
      #   future.get # blocking operation
      #
      # @param app_deployment_uri [UU::OS::UESURI, String] URI of app deployment which will be unshared.
      # @param parameters [Hash] AppDeployment unshare parameters.
      # @option parameters [Array] territories List of territory URIs from where the app deployment will be unshared. Maximum is 1000 territories per command.
      # @option parameters [TrueClass, FalseClass] reportProgress Disable (false) or enable (true) progress reporting
      # (i.e. creation of the appbox activity). By default, reporting is enabled.
      # @return [UU::OS::Lang::Future] Instance of future to use to wait for
      #   the unshare operation to finish. Note that maximum timeout of future
      #   is set to 30 seconds, which is unfortunately insufficient time to finish
      #   unsharing. Therefore timeout error might be expected when +Future.get+
      #   is invoked immediately after the command invocation. It is however
      #   possible to wait repeatedly using the same instance of the returned
      #   +Future+ object.
      def self.unshare(app_deployment_uri, parameters = {})
        self.new(UU::OS::Security::Session.current_session).unshare(app_deployment_uri, parameters)
      end

      # Returns list of application shares.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   list = app_deployment.get_app_share_list("ues:DEV1234-BT:PROD:CDS.GB", pageSize: 30, pageIndex: 2)
      #
      # @example Response
      #   {
      #     :totalSize => 1300,
      #     :criteria => {
      #       :pageIndex => 2,
      #       :pageSize => 30
      #     },
      #     :pageEntries => [
      #       {
      #         :uri => "ues:MT_123:MT_123/CDS_GB",
      #         :code => "MT_123/CDS_GB",
      #         :urlPath => "/cds-gb",
      #         :appDeploymentUri => "ues:MT_123:POOL:CDS.GB",
      #         :territoryUri => "ues:MT_123:MT_123"
      #       }
      #       # ...
      #     ]
      #   }
      #
      # @param app_deployment_uri [UU::OS::UESURI, String] URI of app deployment to list app shares for.
      # @param parameters [Hash] App share list parameters.
      # @option parameters [Fixnum] pageSize Size of one page of records. Default is 1000.
      # @option parameters [Fixnum] pageIndex Page index where 0 means first page, 1 second and so on. Default is 0.
      # @return [Hash<Symbol, Object>] Paged result set.
      # @see .get_app_share_list
      def get_app_share_list(app_deployment_uri, parameters = {})
        @client.invoke(:getAppShareList, app_deployment_uri, parameters: parameters)
      end

      # Returns list of application shares. This method is a convenience shortcut for the {#get_app_share_list} instance
      # method.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   list = UU::C3::AppDeployment.get_app_share_list("ues:DEV1234-BT:PROD:CDS.GB", pageSize: 30, pageIndex: 2)
      #
      # @param app_deployment_uri [UU::OS::UESURI, String] URI of app deployment to list app shares for.
      # @param parameters [Hash] App share list parameters.
      # @option parameters [Fixnum] pageSize Size of one page of records. Default is 1000.
      # @option parameters [Fixnum] pageIndex Page index where 0 means first page, 1 second and so on. Default is 0.
      # @return [Hash<Symbol, Object>] Paged result set.
      def self.get_app_share_list(app_deployment_uri, parameters = {})
        self.new(UU::OS::Security::Session.current_session).get_app_share_list(app_deployment_uri, parameters)
      end

      # Returns attributes of app deployment identified by either its UESURI or
      # combination of TID (Territory ID) and ASID (AppServer ID) associated
      # with that app deployment.
      #
      # @example Get attributes by app deployment UESURI
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   attrs = app_deployment.get_attributes("ues:[123]:[56a3b8ec69a1b474a0000086]:[56a3b8ec69a1b474a0000089]")
      #
      # @example Get attributes by ASID
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   attrs = app_deployment.get_attributes("123-35830958093")
      #
      # @param app_deployment_uri_or_tid_asid [UU::OS::UESURI, String] UESURI
      #   of app deployment or TID+ASID (concatenated with a dash - e.g. "1234-5678").
      # @return [Hash<Symbol, Object>] App deployment attributes.
      def get_attributes(app_deployment_uri_or_tid_asid)
        return @client.invoke(:getAttributes, app_deployment_uri_or_tid_asid)
      end

      # Returns attributes of app deployment identified by either its UESURI or
      # combination of TID (Territory ID) and ASID (AppServer ID) associated
      # with that app deployment. This method is a convenience shortcut for the
      # {#get_attributes} method.
      #
      # @example Get attributes by app deployment UESURI
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.get_attributes("ues:[123]:[56a3b8ec69a1b474a0000086]:[56a3b8ec69a1b474a0000089]")
      #
      # @example Get attributes by ASID
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.get_attributes("123-35830958093")
      #
      # @param app_deployment_uri_or_tid_asid [UU::OS::UESURI, String] UESURI
      #   of app deployment or TID+ASID (concatenated with a dash - e.g. "1234-5678"
      #   where "1234" is TID and "5678" is ASID).
      # @return [Hash<Symbol, Object>] App deployment attributes.
      def self.get_attributes(app_deployment_uri_or_tid_asid)
        return self.new(UU::OS::Security::Session.current_session).get_attributes(app_deployment_uri_or_tid_asid)
      end

      # Modifies the count of nodes on which the AppDeployment is deployed.
      # You can scale out or scale in the entire AppDeployment (all NodeSets)
      # or just the specified NodeSets.
      #
      # @note This command is experimental. And it may not be available for use in all uuCloud regions.
      #
      # @example Scale all NodeSets of a particular AppDeployment to the same count.
      #   require "uu_c3"
      #   app_deployment_uri = "ues:TER:POOL:APP_CODE"
      #   node_count = 3
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   app_deployment.scale(app_deployment_uri, nodeCount: node_count)
      #
      # @example Scale the specified NodeSets of a particular AppDeployment.
      #   require "uu_c3"
      #   app_deployment_uri = "ues:TER:POOL:APP_CODE"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   app_deployment.scale(app_deployment_uri,
      #     nodeSets: [
      #       {
      #         nodeSetUri: "ues:NODE:SET1",
      #         nodeCount: 2
      #       },
      #       {
      #         nodeSetUri: "ues:NODE:SET2",
      #         nodeCount: 5
      #       }
      #     ]
      #   )
      #
      # @param app_deployment_uri [UU::OS::UESURI, String] URI of AppDeployment to scale.
      # @param parameters [Hash] Scale parameters.
      # @option parameters [Fixnum] nodeCount Number of nodes for all NodeSets of the AppDeployment. If +nodeCount+ is
      #   specified, +nodeSets+ must not be specified.
      # @option parameters [Array<Hash>] nodeSets Parameters of the NodeSets to scale. If +nodeSets+ is
      #   specified, +nodeCount+ must not be specified. Supported keys and value types:
      #   * +nodeSetUri+ (UU::OS::UESURI, String) URI of the NodeSet to scale. Required.
      #   * +nodeCount+ (Fixnum) Number of nodes to scale the NodeSet to. Required.
      # @option parameters [UU::C3::AppDeploymentSlot] targetSlot Slot of AppDeployment. Default value is 'PRODUCTION'.
      # @return [UU::OS::Lang::Future] Instance of future to use to wait for
      #   the scale operation to finish. Note that maximum timeout of future
      #   is set to 30 seconds, which may not be sufficient time to finish
      #   scaling in all cases. Therefore timeout error might be expected when +Future.get+
      #   is invoked immediately after the command invocation. It is however
      #   possible to wait repeatedly using the same instance of the returned
      #   +Future+ object.
      # @see .scale
      def scale(app_deployment_uri, parameters = {})
        @client.invoke(:scale, app_deployment_uri, parameters: parameters)
      end

      # Modifies the count of nodes on which the AppDeployment is deployed.
      # This method is a convenience shortcut for the {#scale} method.
      #
      # @note This command is experimental. And it may not be available for use in all uuCloud regions.
      #
      # @example Scale all NodeSets of a particular AppDeployment to the same count.
      #   require "uu_c3"
      #   app_deployment_uri = "ues:TER:POOL:APP_CODE"
      #   node_count = 3
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #   UU::C3::AppDeployment.scale(app_deployment_uri, nodeCount: node_count)
      #
      # @example Scale the specified NodeSets of a particular AppDeployment.
      #   require "uu_c3"
      #   app_deployment_uri = "ues:TER:POOL:APP_CODE"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   UU::C3::AppDeployment.scale(app_deployment_uri,
      #     nodeSets: [
      #       {
      #         nodeSetUri: "ues:NODE:SET1",
      #         nodeCount: 2
      #       },
      #       {
      #         nodeSetUri: "ues:NODE:SET2",
      #         nodeCount: 5
      #       }
      #     ]
      #   )
      #
      # @param app_deployment_uri [UU::OS::UESURI, String] URI of AppDeployment to scale.
      # @param parameters [Hash] Scale parameters.
      # @option parameters [Fixnum] nodeCount Number of nodes for all NodeSets of the AppDeployment. If +nodeCount+ is
      #   specified, +nodeSets+ must not be specified.
      # @option parameters [Array<Hash>] nodeSets Parameters of the NodeSets to scale. If +nodeSets+ is
      #   specified, +nodeCount+ must not be specified. Supported keys and value types:
      #   * +nodeSetUri+ (UU::OS::UESURI, String) URI of the NodeSet to scale. Required.
      #   * +nodeCount+ (Fixnum) Number of nodes to scale the NodeSet to. Required.
      # @option parameters [UU::C3::AppDeploymentSlot] targetSlot Slot of AppDeployment. Default value is 'PRODUCTION'.
      # @return [UU::OS::Lang::Future] Instance of future to use to wait for
      #   the scale operation to finish. Note that maximum timeout of future
      #   is set to 30 seconds, which may not be sufficient time to finish
      #   scaling in all cases. Therefore timeout error might be expected when +Future.get+
      #   is invoked immediately after the command invocation. It is however
      #   possible to wait repeatedly using the same instance of the returned
      #   +Future+ object.
      def self.scale(app_deployment_uri, parameters = {})
        self.new(UU::OS::Security::Session.current_session).scale(app_deployment_uri, parameters)
      end

      # Returns effective environment configuration for nodes deployed in the
      # specified resource pool.
      #
      # "Effective" means that the values are possibly merged from multiple
      # sources (cloud, region, resource group, resource pool) and the values
      # are taken according to the level of granularity of those sources.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   effective_env = app_deployment.get_effective_env("ues:DEV1234-BT:DEV")
      #
      # @example Response
      #   {
      #    "uuClientProperties": {
      #      "source": "Cloud",
      #      "sourceUri": "ues:OT-BT[91343612732354324]:PLUS4U_CLOUD[31cs2238f1428au1b19n513n]",
      #      "value": {
      #        "uu.server-url": "https://api.plus4u.net",
      #        "uu.c3.base_url": "https://sys.plus4u.net",
      #        "uu.c3.append_routing_data": "false",
      #        "uu.imagecatalog.server-path-prefix": "/c00",
      #        "uu.objectstore.server-path-prefix": "/c00",
      #        "uu.operation.registry.base_uri": "https://sys.plus4u.net",
      #        "uu.operation.registry.append_routing_data": "false"
      #      }
      #    },
      #    "gemrc": {
      #      "source": "Cloud",
      #      "sourceUri": "ues:OT-BT[91343612732354324]:PLUS4U_CLOUD[31cs2238f1428au1b19n513n]",
      #      "value": {
      #        ":backtrace": "false",
      #        ":benchmark": "false",
      #        ":bulk_trashold": "100",
      #        ":sources": [
      #          "http://rubygems.org",
      #          "http://gems.plus4u.net"
      #        ],
      #        ":update_sources": "true",
      #        ":verbose": "false",
      #        "gem": "--env-shebang --no-doc --no-ri --no-document"
      #      }
      #    }
      #   }
      #
      #
      # @param resource_pool_uri [UU::OS::UESURI, String] URI of a resource pool.
      # @return [Hash] Effective environment configuration.
      #
      #   * +source+ - determines level of granularity of the source
      #   * +sourceUri+ - UESURI of the particular source the value was taken from
      #   * +value+ - the actual value
      def get_effective_env(resource_pool_uri)
        @client.invoke(:getEffectiveEnv, resource_pool_uri, parameters: {})
      end

      # Returns effective environment configuration for nodes deployed in the
      # specified resource pool.  This method is a convenience shortcut for the
      # {#get_effective_env} method.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   effective_env = UU::C3::AppDeployment.get_effective_env("ues:DEV1234-BT:DEV")
      #
      # @param resource_pool_uri [UU::OS::UESURI, String] URI of a resource pool.
      # @return [Hash] Effective environment configuration.
      def self.get_effective_env(resource_pool_uri)
        self.new(UU::OS::Security::Session.current_session).get_effective_env(resource_pool_uri)
      end

      # @note If you are not sure what this command mean. Take some information about 'Blue/Green deployment strategy' before use.
      # @note Make sure you really want to do this action!
      #   Check if your 'BETA' slot configuration is really configured to be used on 'PRODUCTION' slot.
      # @note If your production application was scaled, remember that this command switch scaled nodes to 'BETA' slot and production will be scaled as 'BETA' before this action.
      # @note This command cannot be called if there is a non-+nil+ +slotMapping+ attribute set on a corresponding ResourceGroup.
      # Swaps between 'PRODUCTION' and 'BETA' slots.
      # This will have impact on AppDeployments accessible on concrete slots(gateways for slots will be changed).
      # This command enables to use newest version of application with zero-downtime to switch.
      # Can be used to rollback previous stable version or test on production environment.
      # @param app_deployment_uri [UU::OS::UESURI, String] URI of application.
      # @return [UU::OS::Lang::Future] Future instance. Make sure than asynchronous part ends successfully.
      def swap(app_deployment_uri)
        @client.invoke(:swap, app_deployment_uri, parameters: {})
      end

      # @note If you are not sure what this command mean. Take some information about 'Blue/Green deployment strategy' before use.
      # @note Make sure you really want to do this action!
      #   Check if your 'BETA' slot configuration is really configured to be used on 'PRODUCTION' slot.
      # @note If your production application was scaled, remember that this command switch scaled nodes to 'BETA' slot and production will be scaled as 'BETA' before this action.
      # @note This command cannot be called if there is a non-+nil+ +slotMapping+ attribute set on a corresponding ResourceGroup.
      # Swaps between 'PRODUCTION' and 'BETA' slots.
      # This will have impact on AppDeployments accessible on concrete slots(gateways for slots will be changed).
      # This command enables to use newest version of application with zero-downtime to switch.
      # Can be used to rollback previous stable version or test on production environment.
      # @param app_deployment_uri [UU::OS::UESURI, String] URI of application.
      # @return [UU::OS::Lang::Future] Future instance. Make sure than asynchronous part ends successfully.
      def self.swap(app_deployment_uri)
        self.new(UU::OS::Security::Session.current_session).swap(app_deployment_uri)
      end

      # Returns list of supported runtime stacks for the specified resource pool.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   app_deployment = UU::C3::AppDeployment.new(UU::OS::Security::Session.current_session)
      #   supported_runtime_stacks = app_deployment.get_supported_runtime_stack_list("ues:DEV1234-BT:DEV")
      #
      # @example Response
      #   {
      #     :totalSize => 3,
      #     :criteria => {
      #       :pageIndex => 0,
      #       :pageSize => 1000
      #     },
      #     :pageEntries => [
      #       {
      #         :name => "RuntimeStack for uuAppg01 with Ruby 2.4",
      #         :code => "UU.APPG01/APPSERVER_RUBY_STACK-2.0",
      #         :uri => "ues:OT-BT[99923616732452117]:UU.APPG01/APPSERVER_RUBY_STACK-2.0[5a213123cf37769a53bf9f52]"
      #       },
      #       {
      #         :name => "RuntimeStack for uuAppg01 with Java",
      #         :code => "UU.APPG01/APPSERVER_JAVA_STACK-1.0",
      #         :uri => "ues:OT-BT[99923616732452117]:UU.APPG01/APPSERVER_JAVA_STACK-1.0[5a213124d2e12b7745e79065]"
      #       },
      #       {
      #         :name => "RuntimeStack for uuAppg01 with Node.js",
      #         :code => "UU.APPG01/APPSERVER_NODEJS_STACK-1.0",
      #         :uri: => "ues:OT-BT[99923616732452117]:UU.APPG01/APPSERVER_NODEJS_STACK-1.0[5a213127cf37769a53bf9f54]"
      #       }
      #     ]
      #   }
      #
      # @param resource_pool_uri [UU::OS::UESURI, String] URI of resource pool for which to list supported runtime stacks.
      # @return [Hash<Symbol, Object>] Paged result set.
      def get_supported_runtime_stack_list(resource_pool_uri, criteria = {})
        @client.invoke(:getSupportedRuntimeStackList, resource_pool_uri, parameters: criteria)
      end

      # Returns list of supported runtime stacks for the specified resource pool. This method is
      # a convenience shortcut for the {#get_supported_runtime_stack_list} instance method.
      #
      # @example
      #   require "uu_c3"
      #
      #   UU::OS::Security::Session.login("app-deployer-passwd")
      #
      #   supported_runtime_stacks = UU::C3::AppDeployment.get_supported_runtime_stack_list("ues:DEV1234-BT:DEV")
      #
      # @param resource_pool_uri [UU::OS::UESURI, String] URI of resource pool for which to list supported runtime stacks.
      # @return [Hash<Symbol, Object>] Paged result set.
      def self.get_supported_runtime_stack_list(resource_pool_uri, criteria = {})
        self.new(UU::OS::Security::Session.current_session).get_supported_runtime_stack_list(resource_pool_uri, criteria)
      end
    end

    # @private
    class Crane < AppDeployment
      def initialize(session = nil)
        puts "Class UU::C3::Crane is deprecated - use UU::C3::AppDeployment instead."
        super
      end
    end

  end
end

