require_relative 'cloud_runtime'
require 'uu/os/cmd/command_client'
require 'uu/os/security/session'

module UU
  module CloudRuntime

    class Gateway < ::UU::CloudRuntime::RuntimeBaseClass

      # Remote service path.
      # @private
      PATH = '/uu-cloud-runtime/Gateway'
      _init_command_client(PATH)

      # Creates a new instance of +Gateway+.
      #
      # @param session [UU::OS::Security::Session] Instance of session representing authenticated user
      #   (can be obtained with {UU::OS::Security::Session#current_session} method).
      def initialize(session = nil)
        @client = UU::OS::CMD::CommandClient.new(PATH, session)
      end

      # Resets configuration for a single gateway.
      #
      # @note Each +slot+ has its own gateway!
      #
      # @example
      #   require "uu_c3"
      #   UU::OS::Security::Session.login("app_credentials")
      #   CLOUD_URI = UU::OS::UESURI.new("ues:OT-BT:PLUS4U_CLOUD")
      #   UU::CloudRuntime::Gateway.reset_configuration(CLOUD_URI, alias: "UUOS9", targetSlot: "PRODUCTION")
      #
      # @param cloud_uri [UU::OS::UESURI, String] Cloud URI.
      # @param parameters [Hash] Parameters.
      # @option parameters [String] alias Gateway alias.
      # @option parameters [String] targetSlot Gateway slot.
      #
      # @return [UU::OS::Lang::Future] Instance of future to use to wait for the operation to finish.
      #
      def reset_configuration(cloud_uri, parameters = {})
        @client.invoke(:resetConfiguration, cloud_uri, parameters: parameters)
      end

      # Resets configuration for a single gateway.
      # This method is a convenience shortcut for the {#reset_configuration} instance method.
      def self.reset_configuration(cloud_uri, parameters = {})
        self.new(UU::OS::Security::Session.current_session).reset_configuration(cloud_uri, parameters)
      end

    end
  end
end
